/* 
 * Mobile browser shell
 */
/* APIs:
	nsIXULBrowserWindow: setJSStatus, setOverLink
	nsIWebProgressListener: state, location, status, security listener
	nsIBrowserDOMWindow: The C++ source has access to browser script through this
				openURI, isTabContentWindow
*/
const nsIBrowserDOMWindow		= Components.interfaces.nsIBrowserDOMWindow;
const nsIDOMChromeWindow		= Components.interfaces.nsIDOMChromeWindow;
const nsIDOMWindow			 	= Components.interfaces.nsIDOMWindow;
const nsIDocShellHistory		= Components.interfaces.nsIDocShellHistory;
const nsIDocShellTreeItem		= Components.interfaces.nsIDocShellTreeItem;
const nsIInterfaceRequestor		= Components.interfaces.nsIInterfaceRequestor;
const nsISHistory				= Components.interfaces.nsISHistory;
const nsISupportsWeakReference 	= Components.interfaces.nsISupportsWeakReference;
const nsIWebNavigation		 	= Components.interfaces.nsIWebNavigation;
const nsIXULBrowserWindow		= Components.interfaces.nsIXULBrowserWindow;
const nsIWebProgress			= Components.interfaces.nsIWebProgress;
const nsIWebProgressListener	= Components.interfaces.nsIWebProgressListener;
const nsIWebProgressListener2	= Components.interfaces.nsIWebProgressListener2;
const nsIXULWindow			 	= Components.interfaces.nsIXULWindow;
const nsIConsoleService			= Components.interfaces.nsIConsoleService;
const nsIObserverService		= Components.interfaces.nsIObserverService;

function dump2() {
	for (var i=0; i<arguments.length; i++)
		dump(arguments[i] + " ");	
	if (arguments.length > 0)
		dump("\n");
}

var gShell = null;
function getShell() {
	if (gShell == null)
		gShell = new TouchShell();
	return gShell;
}

var TouchShell = function() {}

var Macros = {
	getObserverService: function() {
		return Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
	}
}
//
// TouchShell: mobile browser shell
// 
TouchShell.prototype = {
	init: function () 
	{
		try {
			var appCore = Components.classes["@mozilla.org/appshell/component/browser/instance;1"].createInstance(Components.interfaces.nsIBrowserInstance);
			appCore.setWebShellWindow(window);

			this.mouseInterval = 0;
			var browser = getChrome().createBrowser(true);

			// Lots of boilerplate application initialization code from here on
			var xulBrowserWindow = new nsXulBrowserWindow();
			window.QueryInterface(nsIInterfaceRequestor)
				.getInterface(nsIWebNavigation)
				.QueryInterface(nsIDocShellTreeItem).treeOwner
				.QueryInterface(nsIInterfaceRequestor)
				.getInterface(nsIXULWindow)
				.XULBrowserWindow = xulBrowserWindow;
			window.QueryInterface(nsIDOMChromeWindow).browserDOMWindow = new nsBrowserAccess();

			// one-time only initialization. See comments from original at browser/base/content/browser.js
			browser.webNavigation.sessionHistory = 
				Components.classes["@mozilla.org/browser/shistory;1"].createInstance(nsISHistory);
			browser.docShell.QueryInterface(nsIDocShellHistory).useGlobalHistory = true;
			
			var prefs = Components.classes['@mozilla.org/preferences-service;1'].getService(Components.interfaces.nsIPrefBranch);
			this.loadURI(prefs.getCharPref("browser.startup.homepage"));
		} 
		catch (e) {
			alert(e);
			throw(e);
		}
	},
	loadURI: function(uri)
	{
		if (!uri) return;
		getChrome().prepareForLoad();
		getChrome().topBrowser.webNavigation.loadURI(uri, nsIWebNavigation.LOAD_FLAGS_ALLOW_THIRD_PARTY_FIXUP, null, null, null);	
	},
	goBack: function()
	{
		var browser = getChrome().topBrowser;
		if (browser.webNavigation.canGoBack)
			browser.webNavigation.goBack(); 
	},
	goForward: function()
	{
		var browser = getChrome().topBrowser;
		if (browser.webNavigation.canGoForward)
			browser.webNavigation.goForward();
	},
	stopLoading: function()
	{
		getChrome().topBrowser.webNavigation.stop(nsIWebNavigation.STOP_ALL);
	},
	zoomOut: function()
	{
		getChrome().topBrowser.zoomController.scale = 0.5;
	},
	zoomIn: function()
	{
		getChrome().topBrowser.zoomController.scale = 2;	
	},
	reload: function()
	{
		getChrome().topBrowser.webNavigation.reload(nsIWebNavigation.LOAD_FLAGS_NONE);
	},
	newTab: function()
	{
		alert ("TODO newTab");
			//window.QueryInterface(nsIDOMChromeWindow).browserDOMWindow.openURI(null,null,null,null);
	},
	showApp: function(appId)
	{
		getChrome().showApp(appId);
	},
	addBookmark: function()
	{
		alert("TODO");
	},
	showScrollbar: function(visible)
	{
		getChrome().showScrollbar(getChrome().topBrowser, visible);
	},
	setMouseInterval: function(fn, delay, browser)
	{
		this.clearMouseInterval();
		this.mouseInterval = window.setInterval(function() {
			fn();
			getChrome().adjustScrollbar(browser);
		}, delay);
		this.showScrollbar(true);
		return this.mouseInterval;
	},
	clearMouseInterval: function(intervalId)
	{
		var id = intervalId || this.mouseInterval;
		if (id) {
			this.showScrollbar(false);
			window.clearInterval(id);
		}
		this.mouseInterval = 0;
	}
}

function nsXulBrowserWindow() {}

nsXulBrowserWindow.prototype =
{
	QueryInterface : function(aIID)
	{
		if (aIID.equals(nsIXULBrowserWindow) ||
			aIID.equals(nsISupportsWeakReference) ||
			aIID.equals(nsISupports))
			return this;
		throw Components.results.NS_NOINTERFACE;
	},
	setJSStatus : function(status)
	{ // TODO
	},
	setJSDefaultStatus : function(status)
	{ // TODO
	},
	setOverLink : function(link, b)
	{ // TODO
	}
}

function nsBrowserStatusHandler(browser) {
	this.init(browser);
}

nsBrowserStatusHandler.prototype = 
{
	browser: null, 	// <browser> we are tallying for
	canStop: false,	// any interruptible connections active?
	QueryInterface : function(aIID)
	{
		if (aIID.equals(nsIWebProgressListener) ||
			aIID.equals(nsISupportsWeakReference) ||
			aIID.equals(nsISupports))
			return this;
		throw Components.results.NS_NOINTERFACE;
	},
	init : function(browser)
	{
		this.browser =  browser;
		this.onSecurityChange(null, null, nsIWebProgressListener.STATE_IS_INSECURE);
	},	
	destroy : function()
	{
	},
	startLoad: function()
	{
		this.canStop = true;
		getChrome().updateNetworkStatus(this);
	}, 
	endLoad: function()
	{
		this.canStop = false;
		getChrome().updateNetworkStatus(this);
	},
	onStateChange : function(aWebProgress, aRequest, aStateFlags, aStatus)
	{
		if (aStateFlags & nsIWebProgressListener.STATE_IS_NETWORK) {
		// Certain sites (digg.com), do not give us the STATE_IS_NETWORK/STATE_STOP event
		// still working on debugging this
			if (aStateFlags & nsIWebProgressListener.STATE_START) {
				if(aRequest && aWebProgress.DOMWindow == content)
					this.startLoad();
			}
			else if (aStateFlags & nsIWebProgressListener.STATE_STOP) {
				if (aRequest && aWebProgress.DOMWindow == content) 
					this.endLoad();
			}
		}
	},
	onProgressChange : function(aWebProgress, aRequest, aCurSelfProgress, aMaxSelfProgress, aCurTotalProgress, aMaxTotalProgress)
	{
		this.currentTotalProgress = aCurTotalProgress;
		this.maxTotalProgress	 = aMaxTotalProgress;

		var percentage = aCurTotalProgress/aMaxTotalProgress;
		
		//document.getElementById("progressbar").value = parseInt(percentage*100)+"%";	 
	},
	onLocationChange : function(aWebProgress, aRequest, aLocation)
	{
	},	
	onStatusChange : function(aWebProgress, aRequest, aStatus, aMessage)
	{ // TODO
	},
	onSecurityChange : function(aWebProgress, aRequest, aState)
	{ // TODO
	},
	setDefaultStatus : function(status)
	{ // TODO
	}
}

function nsBrowserAccess() {}

nsBrowserAccess.prototype =
{
	QueryInterface : function(aIID)
	{
		if (aIID.equals(nsIBrowserDOMWindow) ||
			aIID.equals(nsISupports))
			return this;
		throw Components.results.NS_NOINTERFACE;
	},
	
	openURI : function(aURI, aOpener, aWhere, aContext)
	{
		var url = aURI ? aURI.spec : "about:blank";
		var browser = getChrome().createBrowser(true);
		var newWindow	= browser.docShell
								 .QueryInterface(nsIInterfaceRequestor)
								 .getInterface(nsIDOMWindow);
		return newWindow;
	},
	isTabContentWindow : function(win)
	{
		return getChrome().isTabContentWindow(win);
	}
}

var gChrome = null;

function getChrome() {
	if (gChrome == null)
		gChrome = new MobileChrome();
	return gChrome;
} 

function MobileChrome() {
	this.init();
}

// MobileChrome is an umbrella class for chrome manipulation
MobileChrome.prototype = {
	topBrowser: null,	// currently visible browser
	init: function() 
	{
	},
	updateTitle: function() 
	{
		document.getElementById("title").value = this.topBrowser.contentTitle;
		document.getElementById("mburlbar").value = this.topBrowser.currentURI.spec;
	},
	updateNetworkStatus: function(statusHandler)
	{
		if (statusHandler.canStop) {
			document.getElementById("load").image = "image/cancel.png"; 
			document.getElementById("load").command = "cmd_BrowserStop"; 
			document.getElementById("title").value = "Loading...";
		}
		else {
			document.getElementById("load").image = "image/refresh.png"; 
			document.getElementById("load").command = "cmd_BrowserReload";
		}
	},
	prepareForLoad: function()
	{
		document.getElementById('title').setAttribute('hidden','false');
		document.getElementById('urlToolBar').setAttribute('hidden','false');
		document.getElementById('urlToolBar').setAttribute('hidden','false');
		document.getElementById('load').setAttribute('hidden','false');
		document.getElementById('bm').setAttribute('hidden','false');
		document.getElementById('commandDeck').setAttribute('selectedIndex','0');	 
	},
	startLoad: function() 
	{
	},
	initButtons: function()
	{
		document.getElementById('load').setAttribute('hidden','false');
		document.getElementById('title').setAttribute('hidden','false');
		document.getElementById('bm').setAttribute('hidden','false');
		document.getElementById('commandDeck').setAttribute('selectedIndex','0');
	},
	showURLBar: function()
	{
		document.getElementById('title').setAttribute('hidden','true');
		document.getElementById('load').setAttribute('hidden','true');
		document.getElementById('bm').setAttribute('hidden','true');
		document.getElementById('commandDeck').setAttribute('selectedIndex','1');
	},
	toggleUrlToolBar: function()
	{
		var elem=document.getElementById('urlToolBar');
		if (elem.getAttribute('hidden') == "true")
			elem.setAttribute('hidden','false');
		else {
			elem.setAttribute('hidden','true');
		}		
		document.getElementById('load').setAttribute('hidden','false');
		document.getElementById('bm').setAttribute('hidden','false');
		document.getElementById('commandDeck').setAttribute('selectedIndex','0');
	},
	getScrollbarSize: function(browser)
	{
		const barWidth = 4;
		var winSize = { width: browser.boxObject.width, 
			height: browser.boxObject.height,
			x: browser.boxObject.x, 
			y: browser.boxObject.y};
		var docSize = { width: browser.contentWindow.document.documentElement.scrollWidth * browser.markupDocumentViewer.fullZoom,
			height: browser.contentWindow.document.documentElement.scrollHeight * browser.markupDocumentViewer.fullZoom };
		var scrollPos = { x: browser.contentWindow.scrollX * browser.markupDocumentViewer.fullZoom,
						y: browser.contentWindow.scrollY * browser.markupDocumentViewer.fullZoom};
		var scale = winSize.height / docSize.height;
		var horiz = {
			x: winSize.width - 4 - barWidth,
			y: winSize.y + scrollPos.y * scale,
			length: winSize.height * scale - 8
		};
		scale = winSize.width / docSize.width;
		var vert= {
			x: winSize.x + scrollPos.x * scale,
			y: winSize.y + winSize.height - 4 - barWidth,
			length: winSize.width * scale - 8
		}
		return {horizontal: horiz, vertical: vert, barWidth: barWidth};
	},
	adjustScrollbar: function(browser)
	{
		var elX = document.getElementById('scrollX');
		var elY = document.getElementById('scrollY');
		if (elX.state == 'closed')
			return;
		s = this.getScrollbarSize(browser);
		// HACK. moveTo does not work if popup is already open. Instead, we hide and reopen the scrollbar
		elY.hidePopup();
		elY.openPopup(null, "", s.horizontal.x, s.horizontal.y, false, false);
//		elY.moveTo(s.horizontal.x, s.horizontal.y);
		elY.sizeTo(s.barWidth, s.horizontal.length);
		// vertical scrollbar
		elX.hidePopup();
		elX.openPopup(null, "", s.vertical.x, s.vertical.y, false, false);
//		elX.moveTo(s.vertical.x, s.vertical.y);
		elX.sizeTo(s.vertical.length, s.barWidth);
	},
	showScrollbar: function(browser, visible)
	{
		var x = document.getElementById('scrollX');
		var y = document.getElementById('scrollY')
		if (x.state != 'open' && visible) {
			x.openPopup(null, "", 0,0, false, false);
			y.openPopup(null, "", 0,0, false, false);
			this.adjustScrollbar(browser);
		}
		else if (x.state != 'closed' && !visible)
		{
			x.sizeTo(0,0);	// HACK a workaround for a bug where browser size gets squished after a popup is shown.
			y.sizeTo(0,0);
			x.hidePopup();
			y.hidePopup();
		}
	},
	showApp: function(appId)
	{
		var els = document.getElementsByClassName("appbox");
		for (var i=0; i < els.length; i++) {
			els[i].collapsed = els[i].id != appId;
		}
	},
	hideBrowser: function(browser)
	{
		dump2("hiding browser");
		this.showScrollbar(browser, false);
		browser.collapsed = true;
//		browser.setAttribute('collapsed', 1);
	},
	handleEvent: function(e)
	{
		if (! e.type in this)
			dump("MouseController called with unknown event type " + e.type + "\n");
		this[e.type](e);
	},
 	DOMTitleChanged: function(ev)
	{
		if (this.topBrowser == null)
			throw("topBrowser is unexpectedly null");
		if (ev.target == this.topBrowser.contentDocument)
			this.updateTitle();
	},
	pageshow: function(ev)	// Document has been displayed
	{
		var browser = ev.currentTarget;
		browser.contentWindow.scrollbars.visible = false;
		browser.zoomController.reset();
//		if (!ev.persisted)
			browser.zoomController.fitContent(); // zoom content
		browser.mouseController.documentLoaded();
	},
	showBrowser: function(browser)
	{
		var browsers = document.getElementsByTagName('browser');
		for (var i = 0; i < browsers.length; i++)
			if (browsers[i] != browser)
				this.hideBrowser(browsers[i]);
		browser.setAttribute('collapsed', 0);
		this.topBrowser = browser;
	},
	browserId: 1,
	// see addTab method in browser/base/content/tabbrowser.xml for how mozilla does it
	createBrowser: function(doShow, pos)
	{
		var b = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
										'browser');
		b.setAttribute('id', "browser-" + this.browserId++);
		b.setAttribute('flex', '1');
		b.setAttribute('type', 'content-primary');
		b.collapsed = true;
		b.zoomController = new ZoomController(b);
		b.mouseController = new MouseController(b);
		b.addEventListener("DOMTitleChanged", this, true);
		b.addEventListener("pageshow", this, true);

		// insert in proper position
		var contain = document.getElementById('browserbox');
		pos = pos ? pos.nextSibling : null;
		while (pos && pos.nodeType != 1)
			pos = pos.nextSibling;
		contain.insertBefore(b, pos);

		if (doShow)
			this.showBrowser(b);
		// broadcast creation
		var e = document.createEvent("Events");
		e.initEvent("NewTab", true, true);
		document.dispatchEvent(e);

		b.stop();	// stop loading about:blank
		b.addProgressListener(new nsBrowserStatusHandler(b), nsIWebProgress.NOTIFY_ALL);
		return b;
	},
	isTabContentWindow: function(aWindow)
	{
		dump("isTabContentWindow\n");
		var browsers = document.getElementsByTagName('browser');
		for (var i = 0; i < browsers.length; i++)
			if (browsers.item(i).contentWindow == aWindow)
				return true;
		return false;
	}
}


var MouseController = function(browser) {
	this.init(browser);
}

MouseController.prototype = {
	browser: null,	// browser we are attached to
	moveStyle: null, // event handler 
	init: function(browser)
	{
		this.browser = browser;
	},
	documentLoaded: function() 
	{
		this.browser.contentDocument.addEventListener("mousedown", this, true);
		this.browser.contentDocument.addEventListener("mouseup",this, true);
		this.browser.contentDocument.addEventListener("dblclick", this, true);
	},
	handleEvent: function(e) 
	{
		if (! e.type in this)
			dump("MouseController called with unknown event type " + e.type + "\n");
		this[e.type](e);
	},
	mousedown: function(e) 
	{
		if (e.target instanceof HTMLInputElement)
			return;
		this.lastEvent = this.firstEvent = e;
		this.fingerDistance = 100;
		this.mousemove = e.button != 2 ? this.mousePan : this.mouseZoom;
		this.browser.contentDocument.addEventListener("mousemove", this, true);
		getChrome().showScrollbar(this.browser, true);
		e.stopPropagation();
		e.preventDefault();
	},
	mouseup: function(e)
	{
		this.browser.contentDocument.removeEventListener("mousemove", this, true);
		getChrome().showScrollbar(this.browser, false);
		//  cancel link clicks if we've been dragging for a while
		var totalDistance = Math.sqrt(
				Math.pow(this.firstEvent.clientX - e.clientX, 2) +
				Math.pow(this.firstEvent.clientY - e.clientY, 2))
		if (totalDistance > 10)
			e.preventDefault();
		// keep scrolling if there is enough momentum
		if (this.lastEvent 
			&& 'momentum' in this.lastEvent
			&& this.mousemove == this.mousePan) 
		{
			var speedX = this.lastEvent.momentum.x / this.lastEvent.momentum.time;
			var speedY = this.lastEvent.momentum.y / this.lastEvent.momentum.time;
//			dump("speedX " + speedX + " speedY " + speedY + "\n");
			const speedLimit = 6;
			if (Math.abs(speedY) > speedLimit)
					speedY = speedY > 0 ? speedLimit : -speedLimit;
			if (Math.abs(speedX) > speedLimit)
					speedX = speedX > 0 ? speedLimit : -speedLimit;
			var lastCall = Date.now();
			browser = this.browser;
			const speedFactor = 0.1;
			const decayFactor = 0.9995;
			const cutoff = 0.2;
//			dump("####\n");
			var intervalId = getShell().setMouseInterval( function() {
				var elapsed = (Date.now() - lastCall);
				var x = elapsed * speedX * speedFactor, y = elapsed * speedY * speedFactor;
//				dump("elapsed:" + elapsed + "\n");
//				dump("sx: " + speedX + " sy:" + speedY + "\n");
//				dump("x: " + x + " y:" + y + "\n");
				browser.contentWindow.scrollBy( -x, -y);
				var slowdown = Math.pow(decayFactor, elapsed);
//				dump("slowdown: " + slowdown + "\n");
				speedX *= slowdown;
				speedY *= slowdown;
				if (Math.abs(speedX) < cutoff && Math.abs(speedY) < cutoff)
					getShell().clearMouseInterval(intervalId);
			}, 0, this.browser);
		}
	},
	mouseZoom: function(e)
	{
		var deltaX = e.screenX - this.firstEvent.screenX + 100;
		var deltaY = e.screenY - this.firstEvent.screenY;
		var newDist = Math.sqrt(Math.pow(deltaX, 2) + Math.pow(deltaY, 2));
		var scale = newDist / this.fingerDistance;
		if (e.screenX < this.firstEvent.screenX && scale > 1)
			scale = 1 / scale;
		var newZoom = scale * this.browser.markupDocumentViewer.fullZoom;
//		dump("dist: " + newDist + " newZoom: " + newZoom + "\n");
		this.fingerDistance = Math.max(0.1, newDist);
		this.browser.zoomController.scale = newZoom;
		this.lastEvent = e;
		getChrome().adjustScrollbar(this.browser);
		e.stopPropagation();
		e.preventDefault();
	},
	mousePan: function(e)
	{
		if (this.lastEvent)
			e.momentum = { time: Math.max(e.timeStamp - this.lastEvent.timeStamp, 1),
				x: e.clientX - this.lastEvent.clientX,
				y: e.clientY - this.lastEvent.clientY,
			};
		this.browser.contentWindow.scrollBy(
			this.lastEvent.clientX - e.clientX,
			this.lastEvent.clientY - e.clientY);
		this.lastEvent = e;
		getChrome().adjustScrollbar(this.browser);
		e.stopPropagation();
		e.preventDefault();
	},
	dblclick: function(e) 
	{
		// find the target by walking the dom.
		// we want to zoom in on the block elements
		var initTarget = e.explicitOriginalTarget;
		e.preventDefault();
		while (initTarget && initTarget.nodeName != "HTML") {
			var disp = window.getComputedStyle(initTarget, "").getPropertyValue("display");
			if (!disp.match(/(inline)/g)) {
				this.browser.zoomController.toggleZoom(initTarget);
				break;
			}
			else
				initTarget = initTarget.parentNode;
		}
	}
}


function ZoomController(aBrowser) {
	this.browser = aBrowser;
};

// ZoomControler sets browser zoom
// It clamps/
ZoomController.prototype = {
	minScale: 0.1,
	maxScale: 3,
	set scale(s) 
	{
		var clamp = Math.min(this.maxScale, Math.max(this.minScale, s));
		clamp = Math.floor(clamp * 1000) / 1000;	// Round to 3 digits
		if (clamp == this.browser.markupDocumentViewer.fullZoom)
			return;
//		dump("scrollY: " + this.browser.contentWindow.scrollY + "\n");//
//		dump("scaling to: " + clamp + "\n");
		this.browser.markupDocumentViewer.fullZoom = clamp;
//		dump("after scrollY: " + this.browser.contentWindow.scrollY + "\n");
		// if we've zoomed out of the viewport, scroll us back in
		var leftEdge = this.browser.contentWindow.scrollX + this.browser.contentWindow.document.documentElement.clientWidth;
		var scrollX = this.browser.contentWindow.document.documentElement.scrollWidth - leftEdge;
 //	 dump("leftEdge:" + leftEdge + " maxWidth:" + this.browser.contentWindow.document.body.style.maxWidth + "\n");
		if (scrollX < 0)
		{
//			dump("had to scroll\n");
			this.browser.contentWindow.scrollBy( scrollX, 0);
		}
	},
	get scale()
	{
		return this.browser.markupDocumentViewer.fullZoom;
	},
	reset: function() 
	{
		this.minScale = ZoomController.prototype.minScale;
		this.maxScale = ZoomController.prototype.maxScale;
	},
	fitContent: function(animate) 
	{
//		dump("fitContent\n");
		animate = animate || false;
		try {
			var oldScale = this.scale;
			this.scale = 1;		// reset the scale to 1 forces document to preferred size
			var body = this.browser.contentWindow.document.body;
			var html = this.browser.contentWindow.document.documentElement;
			var newScale = this.scale;
			var finalWidth = html.clientWidth;
		}
		catch(e) {
			dump(e + "\n");
			return;
		}
		var prefScrollWidth = Math.max(html.scrollWidth, body.scrollWidth); // empirical hack, no idea why
//dump("Browser width:" + this.browser.boxObject.width + "\n");
//dump("WIDTHS: " + "scroll: " + html.scrollWidth + " client " + html.clientWidth + " final:" + finalWidth + "preferred: " + prefScrollWidth + "\n");
		if (prefScrollWidth > (this.browser.boxObject.width - 10) )	{ 
			// body wider than window, scale id down
			// we substract 10 to compensate for 10 pixel browser left margin
			newScale = (this.browser.boxObject.width ) / prefScrollWidth;
			finalWidth = prefScrollWidth;
//			dump("scaling down to " + finalWidth + " scale: " + newScale + "\n");
		}
		body.style.minWidth = body.style.maxWidth = (finalWidth -20) + "px";
		this.minScale = Math.max(this.minScale, newScale);
		if (animate) {
			this.scale = oldScale;
			var zoomController = this;
			var i = Math.max(Math.abs(this.scale - newScale) / 0.1, 1);
			var scaleStep = (newScale - this.scale) / i;
			var intervalId = getShell().setMouseInterval( function() {
				if (i > 0)
					zoomController.scale += scaleStep;
				else {
					zoomController.scale = newScale;
					getShell().clearMouseInterval(intervalId);
				}
				i -= 1;
			}, 0, this.browser);
		}
		else
			this.scale = newScale;
//	dump("scale: " + newScale + " finalWidth: " + finalWidth + "\n");
	},
	getPagePosition: function (el) 
	{
		// this function is often wrong. We really need browser support.  getBoxObjectFor was promising,
		// but it was buggy when zoomed.
		//		var box = el.ownerDocument.getBoxObjectFor(el);
		//		var retVal = { left: box.x, top: box.y, width: box.width, height: box.height,
		//			screenX: box.screenX, screenY: box.screenY};
		var top = 0, left = 0;
		var retVal = { width: el.offsetWidth, height: el.offsetHeight, y:0, x:0};
		while (el != null) {
			retVal.y += el.offsetTop;
			retVal.x += el.offsetLeft;
			el = el.offsetParent;
		}
//		dump("pagePosition x:" + retVal.x +  " y:" + retVal.y + " width:" + retVal.width + " height:" + retVal.height + "\n");
		return retVal;
	},
	getWindowRect: function() 
	{
		return { x: this.browser.contentWindow.scrollX,
				y: this.browser.contentWindow.scrollY,
				width: this.browser.boxObject.width / this.scale,
				height: this.browser.boxObject.height / this.scale }
	},
	toggleZoom: function(el) 
	{
		if (!el) return;
		var elRect = this.getPagePosition(el);
		var winRect = this.getWindowRect();
		// Element is already revealed if:
		if (elRect.x > winRect.x // left is visible
			&& (elRect.x + elRect.width) < (winRect.x + winRect.width) // right is visible
			&& elRect.y > winRect.y // top is visible
			&& elRect.y < (winRect.y + winRect.height * 0.5) // we are in top half of screen
			&& elRect.width > winRect.width * 0.8 ) {	// element is wider than 80% of screen	
			this.fitContent(true);	// already revealed, zoom out
		}
		else
			this.zoomToElement(el);
	},
	getZoomDestination: function(el) 
	{
		var elRect = this.getPagePosition(el);
		var winRect = this.getWindowRect();
		// destination: slightly bellow the top, and zoomed in so you can see element's surroundings
		var margin = 8;//elRect.width / 12;
		var x = elRect.x;
		var width = elRect.width;
		if (elRect.width + margin * 2 < winRect.width) {
			x = Math.max( x - margin, 0);
			width += 2* margin;
		}
		return { x: x, 
				y: Math.max(0, elRect.y - winRect.height * 0.4), 
				scale: (this.browser.boxObject.width- 20) / width };
	},
	zoomToElement: function(el) 
	{	
		var dest = this.getZoomDestination(el);
		var distance = Math.max(Math.abs(dest.y - this.browser.contentWindow.scrollY), 
								Math.abs(dest.x - this.browser.contentWindow.scrollX));
		var scaleDist = Math.abs(this.scale - dest.scale);
		var stepCount = Math.max(distance/20, scaleDist / 0.1); // step count depends on distance in zoom/scrolling
		var stepCount = Math.min(30, Math.floor(stepCount));
		var zoomController = this;
		var intervalId = getShell().setMouseInterval( function() {	
			dest = zoomController.getZoomDestination(el);
//			dump("stepCount:" + stepCount + "\n"); 
//			dump("zoomin to x:" + dest.x + " y:" + dest.y + " scale: " + dest.scale + "\n");
			if (stepCount <= 0)	{
				zoomController.scale = dest.scale;
				zoomController.browser.contentWindow.scrollTo(dest.x, dest.y);
				getShell().clearMouseInterval(intervalId);				
			}
			else {
				var x = Math.floor(zoomController.browser.contentWindow.scrollX + (dest.x - zoomController.browser.contentWindow.scrollX) / stepCount);
				var y = Math.floor(zoomController.browser.contentWindow.scrollY + (dest.y - zoomController.browser.contentWindow.scrollY) / stepCount);
				var scale = zoomController.scale + (dest.scale - zoomController.scale) / stepCount;
//				dump("slide to x:" + x + " y:" + y + " scale:" + scale + 
//				" scrollY:" + zoomController.browser.contentWindow.scrollY + "\n");
				zoomController.scale= scale;
				zoomController.browser.contentWindow.scrollTo(x, y);
			}
			stepCount -= 1;
		}, 0, this.browser);
	}
};

